/* -*- Mode: C++; tab-width: 3; indent-tabs-mode: t; c-basic-offset: 3 -*- */
/*================================================================
 * 
 * Project : Freyja
 * Author  : Terry 'Mongoose' Hendrix II
 * Website : http://www.westga.edu/~stu7440/
 * Email   : stu7440@westga.edu
 * Object  : Dxf
 * License : No use w/o permission (C)2001 Mongoose
 * Comments: DXF module class
 *
 *
 *           This file was generated using Mongoose's C++ 
 *           template generator script.  <stu7440@westga.edu>
 * 
 *-- History ------------------------------------------------- 
 *
 * 2001.07.07:
 * Mongoose - Created
 =================================================================*/

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include "Dxf.h"


Dxf::Dxf()
{
}


Dxf::~Dxf()
{
}



int dxf_parse_entity(FILE *fp, DXF_Data *dat)
{
  int err;
  char value[DXF_LINE_MAX];
  int code;

#ifdef DXF_DEBUG
   fprintf(stderr, "Parsing ENTITY section...\n");
#endif

  /* Parse the different entities */
  while ((err = dxf_getpair(fp, &code, value, dat)))
  {
    dat->lineno++;

    switch (code)
	  {
      case DXF_CODE_ENTITY:
        if (strstr(value, "ENDSEC"))
          return (TRUE);
        else if (strstr(value, "POLYLINE"))
          dxf_add_polyline(dxf_parse_polyline(fp, dat), dat);
        else if (strstr(value, "LINE"))
          dxf_add_line(dxf_parse_line(fp, dat), dat);
        else if (strstr(value, "3DFACE"))
          dxf_add_3dface(dxf_parse_3dface(fp, dat), dat);
        else if (strstr(value, "VERTEX"))
          dxf_add_vertex(dxf_parse_vertex(fp, dat), dat);
        else if (strstr(value, "ARC"))
          dxf_add_arc(dxf_parse_arc(fp, dat), dat);
        else if (strstr(value, "CIRCLE"))
          dxf_add_circle(dxf_parse_circle(fp, dat), dat);
        else if (strstr(value, "SOLID"))
          dxf_add_solid(dxf_parse_solid(fp, dat), dat);
        else if (strstr(value, "POINT"))
          dxf_add_point(dxf_parse_point(fp, dat), dat);
#ifdef DXF_DEBUG
        else
        {
          cr_strip(value);
          fprintf(stderr,"Unrecognized entity: %s\n",value);
        }
#endif 
        break;
    }
    dat->lineno++;
  }
  if (err)
    return (TRUE);
  else
    return (FALSE);
}

/* A polyline. This is a PITA. AC seems to use polylines as some sort of
 * mutant workhorse.
 */
DXF_Polyline *dxf_parse_polyline(FILE *fp, DXF_Data *dat)
{
  int err;
  char value[DXF_LINE_MAX];
  int code;
  int good = TRUE;
  DXF_Polyline *pline;

  /* Malloc() one */
  pline = malloc(sizeof(DXF_Polyline));
  if (pline == NULL)        /* Couldn't allocate, out of memory! */
    return (NULL);

  bzero(pline,sizeof(DXF_Polyline));

  while (good == TRUE)
  {
    err = dxf_getpair(fp, &code, value, dat);
    if (!err)
      break;

    dat->lineno++;

    switch (code)
    {
      case DXF_CODE_ENTITY:
        if (strstr(value, "SEQEND"))
        {
          good = FALSE;
          break;
        }
        if (strstr(value, "VERTEX")) /* Found a vertex for this pline */
	 {
	    dxf_add_vertex_to_polyline(dxf_parse_vertex(fp, dat), pline);
	    break;
	 }
#ifdef DXF_DEBUG
        if (strstr(value, "FACE"))
          fprintf(stderr, "FOUND Pline FACE\n");
#endif
       else fprintf (stderr,"Fould polyline entity: %s\n",value);
        break;
      case 8:        /* Layer */
        cr_strip(value);
        pline->layer = dxf_add_layer(value, dat);
        break;
      case 30:       /* Elevation */
        pline->elevation = strtod(value, NULL);
        break;
      case 39:       /* Thickness */
        pline->thickness = strtod(value, NULL);
        break;
      case 70:       /* Polyline flags */
        pline->flags = atoi(value);
        break;
      case 71:       /* M count */
        pline->mcount = atoi(value);
        break;
      case 72:       /* N Count */
        pline->ncount = atoi(value);
        break;
      case 73:       /* M Density */
        pline->mdensity = atoi(value);
        break;
      case 74:       /* N Density */
        pline->ndensity = atoi(value);
        break;
      case 210:      /* Extrusion vector X */
        pline->extx = strtod(value, NULL);
        break;
      case 220:      /* Extrusion vector Y */
        pline->exty=strtod(value,NULL);
        break;
      case 230:      /* Extrusion vector Z */
        pline->extz=strtod(value,NULL);
        break;
    }
    dat->lineno++;
  }
  if (!err)
  {
    free(pline);
    return (NULL);
  }
  return (pline);
}

/* A  line. 
 */
DXF_Line *dxf_parse_line(FILE *fp, DXF_Data *dat)
{
  int err;
  char value[DXF_LINE_MAX];
  int code;
  DXF_Line *dline;
  int good = TRUE;

  /* Malloc() one */
  dline = malloc(sizeof(DXF_Line));
  if (dline == NULL)
    return NULL;
   
  bzero (dline,sizeof(DXF_Line));
   
  while (good)
  {
    err = dxf_getpair(fp, &code, value, dat);
    if (!err) break;
    dat->lineno++;

    switch (code)
    {
      case 0:    /* Oops, start of new entity */
        good = FALSE;
        /* Now to undo the last read... oh boy */
        dxf_getpair(fp, NULL, NULL, dat);
        /* That's right, NULL's... force a rewind in getpair() */
        break;
      case 8:    /* Layer */
        cr_strip(value);
        dline->layer = dxf_add_layer(value, dat);
        break;
      case 39:
        dline->thickness = strtod(value, NULL);
        break;
      case 10:
        dline->x1 = strtod(value, NULL);
        break;
      case 20:
        dline->y1 = strtod(value, NULL);
        break;
      case 30:
        dline->z1 = strtod(value, NULL);
        break;
      case 11:
        dline->x2 = strtod(value, NULL);
        break;
      case 21:
        dline->y2 = strtod(value, NULL);
        break;
      case 31:
        dline->z2 = strtod(value, NULL);
        break;
      case 210:  /* Extrusion */
        dline->extx = strtod(value, NULL);
        break;
      case 220:
        dline->exty = strtod(value, NULL);
        break;
      case 230:
        dline->extz = strtod(value, NULL);
        break;
      default:
        break;
    }
    dat->lineno++;
  }
   
  if (!err) /* Couldn't parse it! */
  {
    free(dline);
    return NULL;
  }
  /* Check for degenerate */
  if ((dline->x1==dline->x2) && (dline->y1==dline->y2) &&
      (dline->z1==dline->z2))
  {
    dxf_err("Degenerate line.", dat);
    free(dline);
    return(NULL);
  }

  /* Update the bounding box */
  dxf_dobox(dline->x1, dline->y1, dline->z1, dat);
  dxf_dobox(dline->x2, dline->y2, dline->z2, dat);

  return dline;
}

DXF_Vertex *dxf_parse_vertex(FILE *fp, DXF_Data *dat)
{
  int err;
  char value[DXF_LINE_MAX];
  int code;
  DXF_Vertex *vert;
  int good = TRUE;

  /* Malloc() one */
  vert = malloc(sizeof(DXF_Vertex));
  if (vert == NULL)        /* Couldn't allocate, out of memory! */
    return NULL;

  bzero (vert,sizeof(DXF_Vertex));
   
  while (good)
  {
    err = dxf_getpair(fp, &code, value, dat);
    if (!err) break;

    dat->lineno++;

    switch (code)
    {
      case 0: /* Oops, start of new entity */
        good = FALSE;
        /* Now to undo the last read... oh boy */
        dxf_getpair(fp, NULL, NULL, dat);
        /* That's right, NULL's... force a rewind in getpair() */
        break;
      case 8:        /* Layer */
        cr_strip(value);
        vert->layer = dxf_add_layer(value, dat);
        break;
      case 10:
        vert->x = strtod(value, NULL);
        break;
      case 20:
        vert->y = strtod(value, NULL);
        break;
      case 30:
        vert->z = strtod(value, NULL);
        break;
      case 70: /* Flags */
        vert->flags = atoi(value);
        break;
      case 71: /* Polyface mesh vertex index */
        vert->vert_index1=atoi(value);
        break;
      case 72: /* Polyface mesh vertex index */
        vert->vert_index2=atoi(value);
        break;
      case 73: /* Polyface mesh vertex index */
        vert->vert_index3=atoi(value);
        break;
      case 74: /* Polyface mesh vertex index */
        vert->vert_index4=atoi(value);
        break;
      default:
        break;
    }
    dat->lineno++;
  }

  if (!err) /* Couldn't parse it! */
  {
    free(vert);
    return NULL;
  }

  /* Update the bounding box */
  dxf_dobox(vert->x, vert->y, vert->z,dat);

  return vert;
}

/* Parse a 3d-face... you get the idea by now
 */
DXF_3DFace *dxf_parse_3dface(FILE *fp, DXF_Data *dat)
{
  int err;
  char value[DXF_LINE_MAX];
  int code;
  DXF_3DFace *face;
  int good = TRUE;

  /* Malloc() one */
  face = malloc(sizeof(DXF_3DFace));
  if (face == NULL)        /* Couldn't allocate, out of memory! */
    return NULL;

  bzero (face,sizeof(DXF_3DFace));
   
  while (good)
  {
    err = dxf_getpair(fp, &code, value, dat);
    if (!err) break;

    dat->lineno++;

    switch (code)
    {
      case 0: /* Oops, start of new entity */
        good = FALSE;
        /* Now to undo the last read... oh boy */
        dxf_getpair(fp, NULL, NULL, dat);
        /* That's right, NULL's... force a rewind in getpair() */
        break;
      case 8:        /* Layer */
        cr_strip(value);
        face->layer = dxf_add_layer(value, dat);
        break;
      case 10:   /* X point 1 */
        face->x1 = strtod(value, NULL);
        break;
      case 11:
        face->x2 = strtod(value, NULL);
        break;
      case 12:
        face->x3 = strtod(value, NULL);
        break;
      case 13:
        face->x4 = strtod(value, NULL);
        break;
      case 20:
        face->y1 = strtod(value, NULL);
        break;
      case 21:
        face->y2 = strtod(value, NULL);
        break;
      case 22:
        face->y3 = strtod(value, NULL);
        break;
      case 23:
        face->y4 = strtod(value, NULL);
        break;
      case 30:
        face->z1 = strtod(value, NULL);
        break;
      case 31:
        face->z2 = strtod(value, NULL);
        break;
      case 32:
        face->z3 = strtod(value, NULL);
        break;
      case 33:
        face->z4=strtod(value,NULL);
        break;
    }
    dat->lineno++;
  }

  if (!err) /* Couldn't parse it! */
  {
    free(face);
    return (NULL);
  }

  /* Update the bounding box */
  dxf_dobox(face->x1,face->y1,face->z1,dat);
  dxf_dobox(face->x2,face->y2,face->z2,dat);
  dxf_dobox(face->x3,face->y3,face->z3,dat);
  dxf_dobox(face->x4,face->y4,face->z4,dat);

  return (face);
}

/* Parse an ARC
 */
DXF_Arc *dxf_parse_arc(FILE *fp, DXF_Data *dat)
{
  int err;
  char value[DXF_LINE_MAX];
  int code;
  DXF_Arc *arc;
  int good = TRUE;

  /* Malloc() one */
  arc = malloc(sizeof(DXF_Arc));
  if (arc==NULL)        /* Couldn't allocate, out of memory! */
    return NULL;

  bzero (arc,sizeof(DXF_Arc));
   
  while (good)
  {
    err = dxf_getpair(fp, &code, value, dat);
    if (!err) break;

    dat->lineno++;

    switch (code)
    {
      case 0: /* Oops, start of new entity */
        good = FALSE;
        /* Now to undo the last read... oh boy */
        dxf_getpair(fp, NULL, NULL, dat);
        /* That's right, NULL's... force a rewind in getpair() */
        break;
      case 8:        /* Layer */
        cr_strip(value);
        arc->layer = dxf_add_layer(value, dat);
        break;
      case 10:   /* X point 1 */
        arc->x = strtod(value, NULL);
        break;
      case 20:
        arc->y = strtod(value, NULL);
        break;
      case 30:
        arc->z = strtod(value, NULL);
        break;
      case 40:
        arc->radius = strtod(value, NULL);
        break;
      case 50:
        arc->start_angle = strtod(value, NULL);
        break;
      case 51:
        arc->end_angle = strtod(value, NULL);
        break;
      case 210:
        arc->extx = strtod(value, NULL);
        break;
      case 220:
        arc->exty = strtod(value, NULL);
        break;
      case 230:
        arc->extz = strtod(value, NULL);
        break;
    }
    dat->lineno++;
  }

  if (!err) /* Couldn't parse it! */
  {
    free(arc);
    return (NULL);
  }

  /* Update the bounding box */
  dxf_dobox(arc->x+arc->radius, arc->y+arc->radius, arc->z,dat);
  dxf_dobox(arc->x-arc->radius, arc->y-arc->radius, arc->z,dat);

  return (arc);
}

/* POINT Parser block */
DXF_Point *dxf_parse_point(FILE *fp,DXF_Data *dat)
{
  int err;
  char value[DXF_LINE_MAX];
  int code;
  DXF_Point *point;
  int good = TRUE;

  /* Malloc() one */
  point = malloc(sizeof(DXF_Point));
  if (point == NULL)        /* Couldn't allocate, out of memory! */
    return NULL;

  bzero (point,sizeof(DXF_Point));
   
  while (good)
  {
    err = dxf_getpair(fp, &code, value, dat);
    if (!err) break;

    dat->lineno++;

    switch (code)
    {
      case 0: /* Oops, start of new entity */
        good = FALSE;
        /* Now to undo the last read... oh boy */
        dxf_getpair(fp, NULL, NULL, dat);
        /* That's right, NULL's... force a rewind in getpair() */
        break;
      case 8:       /* Layer */
        cr_strip(value);
        point->layer = dxf_add_layer(value, dat);
        break;
      case 10:
        point->x = strtod(value, NULL);
        break;
      case 20:
        point->y = strtod(value, NULL);
        break;
      case 30:
        point->z = strtod(value, NULL);
        break;
      case 39:
        point->thickness = strtod(value, NULL);
        break;
      case 210: /* Extrusion */
        point->extx = strtod(value, NULL);
        break;
      case 220:
        point->exty = strtod(value, NULL);
        break;
      case 230:
        point->extz = strtod(value, NULL);
        break;
    }
    dat->lineno++;
  }

  if (!err) /* Couldn't parse it! */
  {
    free(point);
    return NULL;
  }

  /* Update the bounding box */
  dxf_dobox(point->x, point->y, point->z, dat);

  return point;
}

/* CIRCLE parser block */
DXF_Circle *dxf_parse_circle(FILE *fp, DXF_Data *dat)
{
  int err;
  char value[DXF_LINE_MAX];
  int code;
  DXF_Circle *circle;
  int good = TRUE;

  /* Malloc() one */
  circle = malloc(sizeof(DXF_Circle));
  if (circle == NULL)        /* Couldn't allocate, out of memory! */
    return (NULL);

  bzero (circle,sizeof(DXF_Circle));
   
  while (good)
  {
    err = dxf_getpair(fp, &code, value, dat);
    if (!err) break;

    dat->lineno++;

    switch (code)
    {
      case 0: /* Oops, start of new entity */
        good = FALSE;
        /* Now to undo the last read... oh boy */
        dxf_getpair(fp, NULL, NULL, dat);
        /* That's right, NULL's... force a rewind in getpair() */
        break;
      case 8:        /* Layer */
        cr_strip(value);
        circle->layer=dxf_add_layer(value,dat);
        break;
      case 10:   /* X point 1 */
        circle->x = strtod(value, NULL);
        break;
      case 20:
        circle->y = strtod(value, NULL);
        break;
      case 30:
        circle->z = strtod(value, NULL);
        break;
      case 40:
        circle->radius = strtod(value, NULL);
        break;
      case 210:
        circle->extx = strtod(value, NULL);
        break;
      case 220:
        circle->exty = strtod(value, NULL);
        break;
      case 230:
        circle->extz = strtod(value, NULL);
        break;
    }
    dat->lineno++;
  }

  if (!err) /* Couldn't parse it! */
  {
    free(circle);
    return NULL;
  }

  /* Update the bounding box */
//   dxf_dobox(arc->x+arc->radius,arc->y+arc->radius,arc->z,dat);
   
  return circle;
}

/* SOLID parsing block */
DXF_Solid *dxf_parse_solid(FILE *fp, DXF_Data *dat)
{
  int err;
  char value[DXF_LINE_MAX];
  int code;
  DXF_Solid *solid;
  int good = TRUE;

  /* Malloc() one */
  solid = malloc(sizeof(DXF_Solid));
  if (solid == NULL)        /* Couldn't allocate, out of memory! */
    return NULL;

  bzero (solid,sizeof(DXF_Solid));
   
  while (good)
  {
    err = dxf_getpair(fp, &code, value, dat);
    if (!err) break;

    dat->lineno++;

    switch (code)
    {
      case 0: /* Oops, start of new entity */
        good = FALSE;
        /* Now to undo the last read... oh boy */
        dxf_getpair(fp, NULL, NULL, dat);
        /* That's right, NULL's... force a rewind in getpair() */
        break;
      case 8:        /* Layer */
        cr_strip(value);
        solid->layer = dxf_add_layer(value, dat);
        break;
      case 10:   /* X point 1 */
        solid->x1 = strtod(value, NULL);
        break;
      case 11:
        solid->x2 = strtod(value, NULL);
        break;
      case 12:
        solid->x3 = strtod(value, NULL);
        break;
      case 13:
        solid->x4 = strtod(value, NULL);
        break;
      case 20:
        solid->y1 = strtod(value, NULL);
        break;
      case 21:
        solid->y2 = strtod(value, NULL);
        break;
      case 22:
        solid->y3 = strtod(value, NULL);
        break;
      case 23:
        solid->y4 = strtod(value, NULL);
        break;
      case 30:
        solid->z1 = strtod(value, NULL);
        break;
      case 31:
        solid->z2 = strtod(value, NULL);
        break;
      case 32:
        solid->z3 = strtod(value, NULL);
        break;
      case 33:
        solid->z4 = strtod(value, NULL);
        break;
      case 39:
        solid->thickness = strtod(value, NULL);
        break;
      case 210:
        solid->extx = strtod(value, NULL);
        break;
      case 220:
        solid->exty = strtod(value, NULL);
        break;
      case 230:
        solid->extz=strtod(value,NULL);
        break;
    }
    dat->lineno++;
  }
  if (!err) /* Couldn't parse it! */
  {
    free(solid);
    return (NULL);
  }

   /* Update the bounding box */
//   dxf_dobox(arc->x+arc->radius,arc->y+arc->radius,arc->z,dat);

  return (solid);
}



/* Open a DFX file stream.
 * This and dfx_close() are the only two functions shared by both ASCII
 * and binary DFX reads.
 */

FILE *dxf_open(char *fname)
{
  FILE *fp;

  fp = fopen(fname, "r");

  return fp;
}

/* Get a line from a DXF file. Return NULL on error.*/
int dxf_getline(FILE *fp, char *line, DXF_Data *dat)
{
  if (dat->type == DXF_ASCII)
    return (dxf_getaline(fp, line));
  else
  {
    fprintf(stderr, "Binary parsing not implemented yet\n");
    return FALSE;
  }
}

/* Get a line from a DXF ASCII file.
 * returns positive on success, FALSE otherwise.
 */
int dxf_getaline(FILE *fp, char *line)
{
  char *err;
  err = fgets(line, DXF_LINE_MAX, fp);

  if (err==NULL)
    return FALSE;
  else
    return TRUE;
}

/* Close a DFX file stream.
 * This and dfx_open() are the only two functions shared between ASCII
 * and binary DFX reads.
 */
int dxf_close(FILE *fp)
{
  return (fclose(fp));
}

/* Determine the type of error for a DXF ASCII file read.
 * Since fgets() doesn't discern between EOF and other stream errors, 
 * we figgure it out here.
 * returns either EOF (for end of file), or the value of the global errno.
 */
int dxf_error(FILE *fp)
{
  if (feof(fp))
    return EOF;
  else
    return errno;
}

/* Get a group code and value pair (wrapper)
 */
int dxf_getpair(FILE *fp, int *code, char *value, DXF_Data *dat)
{
  int good;
   
  if (dat->type == DXF_ASCII)
  {
    good = dxf_getapair(fp, code, value);
    if (!good) return (good);
    if ((code==NULL)||(value==NULL))
       return(good);
    while ((*code) == 999)
    {
	    good = dxf_getapair(fp, code, value);
      if (!good) return (good);
    }
    return (good);
  }
  else
    return (dxf_getbpair(fp, code, value));
}

/* Get a group code and value pair. Binary style.
 * Returns FALSE on fail, else positive.
 * Silly but nescessary use of fseek(). May break char streams on some
 * platforms... should maybe use fgetpos() and fsetpos(). If 'code' and 
 * 'value' are NULL when this func is called, the rewind is enabled, and no
 * read is done. (return value is pointless)
 * This routine makes some BOLD assumptions about datatype sizes and
 * endianness. This NEEDS to be corrected.
 */
int dxf_getbpair(FILE *fp,int *code, char *value)
{
  char buf[512];
  unsigned char gc;
  short int tgc;
  double dvalue;

  /* First, get the group code */
  if (!(fread(&gc, 1, 1, fp)))
    return FALSE;

  /* Check to be sure this isn't an extended code */
  if (gc == 255)
  {           /* It is */
    if (!fread (&tgc, 2, 1, fp))
      return (FALSE);
  }
  else
    tgc = (short int)gc;

  if ( ((tgc>=10)&&(tgc<=59)) || ((tgc>=140)&&(tgc<=147)) ||
       ((tgc==210) || (tgc==220) || (tgc==230)))
  {
    /* It's a double, in 8 bytes */
    if (!fread(&dvalue, 8, 1,fp))
      return FALSE;
    sprintf(buf, "%e", dvalue);
    strcpy(value, buf);
    *code = tgc;
    return TRUE;
  }
  return TRUE;
}

/* Get a group code and value pair. ASCII style.
 * Returns FALSE on fail, else positive.
 * Silly but nescessary use of fseek(). May break char streams on some
 * platforms... should maybe use fgetpos() and fsetpos(). If 'code' and 
 * 'value' are NULL when this func is called, the rewind is enabled, and no
 * read is done. (return value is pointless)
 */
int dxf_getapair(FILE *fp, int *code, char *value)
{
  char buf[DXF_LINE_MAX];
  static long fpos=0;
   
  /* Kludgy, but the spec sucks, what can I say?
   * Some entities have an end marker, some don't.
   * for the ones that don't, a rewind func, built in to the read func!
   * (DOH!)
   * If some func just read a pair, realised that they weren't more
   * entity data but rather the begining of a new entity, and thus
   * the end of this entity, they can use this to put the pair back on
   * the "stack" for the calling func to parse.
   */
  if ( (code==NULL) && (value==NULL) )
  {
    fseek(fp, fpos, SEEK_SET);
    return(TRUE);
  }
  /* Okay, no rewind... on with the show */

  /* Set our current pos. before nuking it with a read */
  fpos = ftell(fp);

  if (!dxf_getaline(fp, buf))
    return FALSE;

  *code = atoi(buf);

  if (!dxf_getaline(fp, buf))
    return FALSE;

  strncpy(value, buf, DXF_LINE_MAX);
 
  return TRUE;
}

/* Parse a section
 */
int dxf_parse_section(FILE *fp,DXF_Data *dat)
{
  int code;
  char value[DXF_LINE_MAX];
  int err;
  if (!dxf_getpair(fp, &code, value, dat))
    return FALSE;
  if (code != DXF_CODE_NAME)
    return FALSE;

  if (strstr(value, "ENTITIES"))
  {
    err = dxf_parse_entity(fp, dat);
    if (!err)
      return FALSE;
    else
      return TRUE;
  }

#ifdef DXF_DEBUG
  cr_strip(value);
  fprintf(stderr, "%s section not yet implemented\n", value);
#endif
  /* For now, seek EOS */
  while ((err = dxf_getpair(fp, &code, value, dat)))
  {
    dat->lineno++;
    if (code != DXF_CODE_ENTITY)
      continue;
    if (!strstr(value, "ENDSEC"))
      continue;
    break;
  }
  if (err)
    return TRUE;
  else
    return FALSE;
}

/* This tries to determine if a DXF file is ASCII or binary.
 * We don't deal w/ binary right now, but we will.
 * this func. uses rewind()... if this breaks anything for anyone, please
 * let me know.
 */
void Dxf::Check(FILE *f)
{
  char header[22];


  fread(header, 22, 1, f);
  header[21] = 0;

  if (strncmp("AutoCAD Binary DXF\n\r", head, 20) == 0)
    _type = DXF_BINARY;
  else if (strncmp("AutoCAD Binary DXF\n\r", head, 20) == 0)
    _type = DXF_ASCII;
  else
    _type = DXF_ERROR;
}


/* This reads a NULL terminated string from fp */
int Dxf::ReadString(FILE *f)
{
  char c;


  c = fgetc(f);

  while ((c != EOF)&&(c != 0))
    *string++ = c;

  if (c == EOF)
    return -1;

  *string = 0;

  return 0;
}


int Dxf::Load(char *filename)
{
  FILE *f;
  char buf[256];
  int code;
  char value[DXF_LINE_MAX];
  int err;


  f = dxf_open(filename);

  if (!f)
  {
	  perror("Dxf::Load> Failed to load file\n");
	  return -1;
  }

  dxf_determine_type(fp, dat);

#ifdef DXF_DEBUG
  if (dat->type == DXF_ASCII)
    fprintf(stderr, "ASCII DXF file.\n");
  else if (dat->type == DXF_BINARY)
    fprintf(stderr, "Binary DXF File.\n");
  else
#endif
  if (dat->type == DXF_ERROR)
  {
    perror("Couldn't determine DXF file type");
    dxf_close(fp);
    return NULL;
  }
  if (dat->type == DXF_BINARY)
  {
    dxf_err("Binary DXF files not implemented yet.\n", dat);
    dxf_close(fp);
    return NULL;
  }
  /* ASCII Only for now... */
  while ((err = dxf_getpair(fp, &code, value, dat)))
  {
    dat->lineno++; /* First line of the pair */
    /* We're expecting section headers here, and only section headers! */
    if (code != DXF_CODE_ENTITY)
    {
      /* Damn.. not a section header.. not even an entity type! */
      dxf_err("Didn't find expected entity type.", dat);
      dxf_close(fp);
      return NULL;
    }
    /* Up the line count */
    dat->lineno++;

    /* Now to check the entity type.... */
    if (strstr(value, "EOF"))
    {
      /* Found the proper End Of File marker... bail out */
      break;
    }
    /* Not EOF, so it'd better be a SECTION or we're DOOMED! :) */
    if (!strstr(value, "SECTION"))
    {
      /* Doomed, I tell ya.... */
      dxf_err("Didn't find expected SECTION entity.", dat);
      dxf_close(fp);
      return NULL;
    }
    if (!dxf_parse_section(fp, dat))
    {
      dxf_err("Error parsing section", dat);
      dxf_close(fp);
      return NULL;
	  }
  }
  if (err == FALSE)
  {
    dxf_err("Some kind of error occured.", dat);
  }

#ifdef DXF_DEBUG
  else
    fprintf(stderr, "DXF File parsed successfully.\n");

  fprintf(stderr, "BBox:  x1:% 9.5f y1:% 9.5f z1:% 9.5f\n",
          dat->bbox_xmin, dat->bbox_ymin, dat->bbox_zmin);
  fprintf(stderr, "       x2:% 9.5f y2:% 9.5f z2:% 9.5f\n",
          dat->bbox_xmax, dat->bbox_ymax, dat->bbox_zmax);
  fprintf(stderr, "\nLayers: %d\n", dat->layer_count);

  fprintf(stderr, "\nTotal Entities: %d\n\n", dat->entities);
  fprintf(stderr, "SOLIDS: %d\n3DFACES: %d\nPOLYLINES: %d\n",
          dat->solid_count, dat->threedface_count, dat->pline_count);
  fprintf(stderr, "VERTICES: %d\nLINES: %d\nARCS: %d\n",
          dat->vert_count, dat->line_count, dat->arc_count);
  fprintf(stderr,"CIRCLES: %d\nPOINTS: %d\n",
          dat->circle_count, dat->point_count);
#endif
  dxf_close(fp);

#ifdef DXF_DEBUG
  dxf_free_all(dat);
   free(dat);
#endif
   
  return dat;
}

/* Simple "print error and line number" routine */     
void dxf_err(char *string, DXF_Data *dat)
{
  fprintf (stderr, "%d: %s\n", dat->lineno, string);
}

/* Nuke trailing CRs and LFs from the end of a string. Not prefixed with
 * "dxf_" because this routine might be useful elsewhere. Let me know if
 * there's a clash
 */
void cr_strip(char *string)
{
  int len=strlen(string)-1;
   
  while ((string[len] == '\n') || (string[len] == '\r'))
  {
    string[len] = 0;
    len = strlen(string-1);
  }
}

/* Resize the bounding-box of this object/scene */
void dxf_dobox(double x, double y, double z, DXF_Data *dat)
{
   if (x > dat->bbox_xmax) dat->bbox_xmax = x;
   if (y > dat->bbox_ymax) dat->bbox_ymax = y;
   if (z > dat->bbox_zmax) dat->bbox_zmax = z;

   if (x < dat->bbox_xmin) dat->bbox_xmin = x;
   if (y < dat->bbox_ymin) dat->bbox_ymin = y;
   if (z < dat->bbox_zmin) dat->bbox_zmin = z;
}

/* Add a polyline to the linked list of polylines in DXF_Data */
int dxf_add_polyline(DXF_Polyline *pline, DXF_Data *dat)
{
  DXF_Polyline *iter;

  /* Up the counts */
  dat->pline_count++;
  dat->entities++;

  /* NULL out the next pointer on new polyline */
  pline->next=NULL;

  /* First case: no plines added yet */

  if (dat->plines==NULL)
  {
    dat->plines = pline;
    return TRUE;
  }

  /* Okay, now to find the end of the chain */
  for (iter=dat->plines ; iter->next ; iter=iter->next);

  /* Now to add the new one */
  iter->next = pline;

  return TRUE;
}

/* Add a 3d-face to the list */
int dxf_add_3dface(DXF_3DFace *face, DXF_Data *dat)
{
  DXF_3DFace *iter;

  /* Up the counts */
  dat->threedface_count++;
  dat->entities++;

  /* NULL out the next pointer on new face */
  face->next = NULL;

  /* First case: no faces added yet */

  if (dat->threedfaces == NULL)
  {
    dat->threedfaces = face;
    return (TRUE);
  }

  /* Okay, now to find the end of the chain */
  for (iter = dat->threedfaces ; iter->next ; iter = iter->next);

  /* Now to add the new one */
  iter->next = face;

  return (TRUE);
}

/* Add a CIRCLE */
int dxf_add_circle(DXF_Circle *circle, DXF_Data *dat)
{
  DXF_Circle *iter;

  /* Up the counts */
  dat->circle_count++;
  dat->entities++;

  /* NULL out the next pointer on new face */
  circle->next = NULL;

  /* First case: no faces added yet */
  if (dat->circles == NULL)
  {
    dat->circles=circle;
    return (TRUE);
  }

  /* Okay, now to find the end of the chain */
  for (iter=dat->circles ; iter->next ; iter=iter->next);

  /* Now to add the new one */
  iter->next = circle;

  return (TRUE);
}

/* Add an ARC */
int dxf_add_arc(DXF_Arc *arc, DXF_Data *dat)
{
  DXF_Arc *iter;

  /* Up the counts */
  dat->arc_count++;
  dat->entities++;

  /* NULL out the next pointer on new face */
  arc->next = NULL;

  /* First case: no faces added yet */
   
  if (dat->arcs == NULL)
  {
    dat->arcs = arc;
    return (TRUE);
  }

  /* Okay, now to find the end of the chain */
  for (iter=dat->arcs ; iter->next ; iter=iter->next);

  /* Now to add the new one */
  iter->next = arc;

  return (TRUE);
}

/* Add SOLID */
int dxf_add_solid(DXF_Solid *solid, DXF_Data *dat)
{
  DXF_Solid *iter;

  /* Up the counts */
  dat->solid_count++;
  dat->entities++; 

  /* NULL out the next pointer on new face */
  solid->next = NULL;

  /* First case: no faces added yet */
   
  if (dat->solids == NULL)
  {
    dat->solids = solid;
    return (TRUE);
  }

  /* Okay, now to find the end of the chain */
  for (iter=dat->solids ; iter->next ; iter=iter->next);

  /* Now to add the new one */
  iter->next = solid;

  return (TRUE);
}

/* Add a solo vertex */
int dxf_add_vertex(DXF_Vertex *vert, DXF_Data *dat)
{
  DXF_Vertex *iter;

  vert->next = NULL;

  /* Up the counts */
  dat->vert_count++;
  dat->entities++;

  if (dat->verts == NULL)
  {
    dat->verts = vert;
    return (TRUE);
  }

  /* Okay, now to find the end of the chain */
  for (iter=dat->verts ; iter->next ; iter=iter->next);

  /* Now to add the new one */
  iter->next = vert;

  return (TRUE);
}

/* Insert a LINE into the list */
int dxf_add_line(DXF_Line *line, DXF_Data *dat)
{
  DXF_Line *iter;
   
  line->next = NULL;

  /* Up the counts */
  dat->line_count++;
  dat->entities++;

  if (dat->lines == NULL)
  {
    dat->lines = line;
    return (TRUE);
  }

  /* Okay, now to find the end of the chain */
  for (iter=dat->lines ; iter->next ; iter=iter->next);

  /* Now to add the new one */
  iter->next = line;

  return (TRUE);
}

/* Insert a new POINT */
int dxf_add_point(DXF_Point *point, DXF_Data *dat)
{
  DXF_Point *iter;

  point->next = NULL;

  /* Up the counts */
  dat->point_count++;
  dat->entities++;

  if (dat->points == NULL)
  {
    dat->points = point;
    return (TRUE);
  }

  /* Okay, now to find the end of the chain */
  for (iter=dat->points ; iter->next ; iter=iter->next);

  /* Now to add the new one */
  iter->next = point;

  return (TRUE);
}

/* Add a VERTEX to the linked list of vertices in a POLYLINE */
int dxf_add_vertex_to_polyline(DXF_Vertex *vertex, DXF_Polyline *pline)
{
  DXF_Vertex *iter;

  /* NULL out new next */
  vertex->next = NULL;

  /* First case */
  if (pline->verts == NULL)
  {
    pline->verts = vertex;
    return (TRUE);
  }

  /* Okay, traverse the chain */
  for (iter=pline->verts ; iter->next ; iter=iter->next);

  /* Plug it in */
  iter->next = vertex;

  return (TRUE);
}

/* Add a layer to the list and return a pointer, or return a pointer
 * to an existing layer string if a match already exists
 */
char *dxf_add_layer(char *name, DXF_Data *dat)
{
  DXF_Layer *iter;
  DXF_Layer *layer;

  layer = malloc(sizeof(DXF_Layer));
  if (layer == NULL)
    return (NULL);

  strcpy(layer->name, name);
  layer->next = NULL;

  /* First case: is the layers pointer empty? If so, life is easy */
  if (dat->layers == NULL)
  {
    dat->layers = layer;
    dat->layer_count++;
    return (layer->name);
  }

  /* Determine if we know about this layer already */
  for (iter=dat->layers ; iter ; iter=iter->next)
  {
    if (!strcmp(iter->name, name))
    {
      free(layer);
      return (iter->name);
    }
  }

  /* Okay, need to insert a new layer object */
  for (iter=dat->layers ; iter->next ; iter=iter->next);
  iter->next = layer;
  dat->layer_count++;
  return (layer->name);
}

/* Free up the entire tree... this should de-allocate ANY allocated mem,
 * since all allocated mem is attatched to DXF_Data somehow.
 */
void dxf_free_all(DXF_Data *dat)
{
  DXF_Vertex *verts, *vold;
  DXF_Line *lines, *lold;
  DXF_3DFace *faces, *fold;
  DXF_Polyline *plines, *pold;
  DXF_Layer *layers, *lsold;
  DXF_Arc *arcs, *aold;
  DXF_Circle *circs, *cold;
  DXF_Solid *sols, *sold;
  DXF_Point *points, *pnold;

  /* Free the solo verts */
  verts = dat->verts;
  while (verts)
  {
    vold = verts;
    verts = verts->next;
    free(vold);
  }

  /* Free the lines */
  lines = dat->lines;
  while (lines)
  {
    lold = lines;
    lines = lines->next;
    free(lold);
  }

  /* Free the faces */
  faces = dat->threedfaces;
  while(faces)
  {
    fold = faces;
    faces = faces->next;
    free(fold);
  }

  /* Free up the polylines */
  plines = dat->plines;
  while (plines)
  {
    verts = plines->verts;
    while (verts)
    {
      vold = verts;
      verts = verts->next;
      free(vold);
    }
    pold = plines;
    plines = plines->next;
    free(pold);
  }

  /* Free the ARCS */
  arcs = dat->arcs;
  while (arcs)
  {
    aold = arcs;
    arcs = arcs->next;
    free(aold);
  }

  /* Free the CIRCLES */
  circs = dat->circles;
  while (circs)
  {
    cold = circs;
    circs = circs->next;
    free(cold);
  }

  /* Free the SOLIDS */
  sols = dat->solids;
  while (sols)
  {
    sold = sols;
    sols = sols->next;
    free(sold);
  }

  /* Free the POINTS */
  points = dat->points;
  while (points)
  {
    pnold = points;
    points = points->next;
    free(pnold);
  }

  /* Free up the layers */
  layers = dat->layers;
  while (layers)
  {
    lsold = layers;
    layers = layers->next;
    free(lsold);
  }
}

#ifndef DXF_STANDALONE
/*****************************************************************************
*  LoadDXF
******************************************************************************/
void LoadDXF(GtkWidget *w, GtkFileSelection *FS)
{
  ObjectStruct *TmpObject;
  GtkWidget *menu_item;
  char *Name;
  FrameStruct *TmpFrame;
  DXF_Data *dat;
  DXF_3DFace *Tmp3dFace;
  DXF_Line *TmpLine;
  DXF_Vertex *TmpVertex;

  Name = g_strdup(gtk_file_selection_get_filename(GTK_FILE_SELECTION(FS)));
 
  dat = dxf_parse(Name);
  
/***************************** INSERT HERE ********************************/
  if (dat == NULL)
    return;
  TmpFrame = NewFrame(Name);

  /* 3DFaces*/
  {
    ObjectStruct *TmpMesh;
    ObjectStruct *FirstMesh = NULL;
    ObjectStruct *TmpObject;

    for (Tmp3dFace = dat->threedfaces ; Tmp3dFace ; Tmp3dFace=Tmp3dFace->next)
    {
      Vector P1, P2, P3;

      for (TmpMesh = FirstMesh ; 
           TmpMesh && strcmp(TmpMesh->Name, Tmp3dFace->layer) ; 
           TmpMesh = TmpMesh->Next)
        ;
      if (TmpMesh == NULL)
      {
        TmpMesh = giram_mesh_new();
        TmpMesh->Next = FirstMesh;
        FirstMesh = TmpMesh;
        if (Tmp3dFace->layer)
          TmpMesh->Name = g_strdup(Tmp3dFace->layer);
        else TmpMesh->Name = g_strdup("Null Layer");
      }

      P1[0] = Tmp3dFace->x1; P1[1] = Tmp3dFace->y1; P1[2] = Tmp3dFace->z1;
      P2[0] = Tmp3dFace->x2; P2[1] = Tmp3dFace->y2; P2[2] = Tmp3dFace->z2;
      P3[0] = Tmp3dFace->x3; P3[1] = Tmp3dFace->y3; P3[2] = Tmp3dFace->z3;
      giram_mesh_add_triangle(TmpMesh, P1, P2, P3);

      if ((Tmp3dFace->x3 != Tmp3dFace->x4) ||
          (Tmp3dFace->y3 != Tmp3dFace->y4) ||
          (Tmp3dFace->z3 != Tmp3dFace->z4))
      {

        P1[0] = Tmp3dFace->x1; P1[1] = Tmp3dFace->y1; P1[2] = Tmp3dFace->z1;
        P2[0] = Tmp3dFace->x4; P2[1] = Tmp3dFace->y4; P2[2] = Tmp3dFace->z4;
        P3[0] = Tmp3dFace->x3; P3[1] = Tmp3dFace->y3; P3[2] = Tmp3dFace->z3;
        giram_mesh_add_triangle(TmpMesh, P1, P2, P3);
      }
    }
    if (TmpFrame->FirstObject)
    {
      for (TmpObject = TmpFrame->FirstObject ; 
           TmpObject->Next ; 
           TmpObject = TmpObject->Next)
        ;
      TmpObject->Next = FirstMesh;
    } else
    {
      TmpFrame->FirstObject = FirstMesh;
    }
  }
  /* Lines*/
  for (TmpLine = dat->lines ; TmpLine ; TmpLine=TmpLine->next)
  {
    ObjectStruct *Cylinder;
    Vector Base, Apex;

    Base[0] = TmpLine->x1; Base[1] = TmpLine->y1; Base[2] = TmpLine->z1;
    Apex[0] = TmpLine->x2; Apex[1] = TmpLine->y2; Apex[2] = TmpLine->z2;
    Cylinder = giram_cylinder_new(Base, Apex, 10e-4);
    Cylinder->Next = TmpFrame->FirstObject;
    TmpFrame->FirstObject = Cylinder;
  }

  /* Vertices */
  for (TmpVertex = dat->verts ; TmpVertex ; TmpVertex=TmpVertex->next)
  {
    ObjectStruct *Sphere;
    Vector Center;

    Center[0]=TmpVertex->x; Center[1]=TmpVertex->y; Center[2]=TmpVertex->z;
    Sphere = giram_sphere_new(Center, 10e-4);
    Sphere->Next = TmpFrame->FirstObject;
    TmpFrame->FirstObject = Sphere;
  }
/**************************************************************************/  
  dxf_free_all(dat);
  free(dat);
  for (TmpObject = TmpFrame->FirstObject ;
       TmpObject ;
       TmpObject = TmpObject->Next)
  {
    giram_object_build_triangle_mesh(TmpObject);
  }
  ComputeCamera(TmpFrame->Camera);
  if (CSGTreeDialog)
  {
    menu_item = gtk_radio_menu_item_new_with_label(CSGTreeMenuGroup,
                                                   TmpFrame->Name);
    gtk_signal_connect(GTK_OBJECT(menu_item), "activate",
                       (GtkSignalFunc)RebuildCSGTree, TmpFrame);
    if (!CSGTreeMenuGroup)
      CSGTreeMenuGroup = gtk_radio_menu_item_group(GTK_RADIO_MENU_ITEM(menu_item));
    gtk_menu_append(GTK_MENU(CSGTreeMenu), menu_item);
    gtk_object_set_data(GTK_OBJECT(menu_item), "CSGMenu", TmpFrame);
    gtk_widget_show(menu_item);
  }
  /* XXX: the created views should be chosen from the 'Preferences' */
  NewView(TmpFrame, ORTHO_XY_CAMERA);
  NewView(TmpFrame, ORTHO_XZ_CAMERA);
  NewView(TmpFrame, ORTHO_ZY_CAMERA);
  NewViewCamera(TmpFrame);
}

/*****************************************************************************
*  InsertDXF
******************************************************************************/
void InsertDXF(GtkWidget *w, GtkFileSelection *FS)
{
  char *Name;
  FrameStruct *TmpFrame;
  FrameViewStruct *TmpView;
  DXF_Data *dat;
  DXF_3DFace *Tmp3dFace;
  DXF_Line *TmpLine;
  DXF_Vertex *TmpVertex;

  Name = g_strdup(gtk_file_selection_get_filename(GTK_FILE_SELECTION(FS)));
 
  dat = dxf_parse(Name);

/***************************** INSERT HERE ********************************/
  if (dat == NULL) 
    return;
  TmpFrame = (FrameStruct *)gtk_object_get_data(GTK_OBJECT(FS), "frame");
 
  /* 3DFaces*/
  {
    ObjectStruct *TmpMesh;
    ObjectStruct *FirstMesh = NULL;
    ObjectStruct *TmpObject;

    for (Tmp3dFace = dat->threedfaces ; Tmp3dFace ; Tmp3dFace=Tmp3dFace->next)
    {
      Vector P1, P2, P3;

      for (TmpMesh = FirstMesh ; 
           TmpMesh && strcmp(TmpMesh->Name, Tmp3dFace->layer) ; 
           TmpMesh = TmpMesh->Next)
        ;
      if (TmpMesh == NULL)
      {
        TmpMesh = giram_mesh_new();
        TmpMesh->Next = FirstMesh;
        FirstMesh = TmpMesh;
        if (Tmp3dFace->layer)
          TmpMesh->Name = g_strdup(Tmp3dFace->layer);
        else TmpMesh->Name = g_strdup("Null Layer");
      }

      P1[0] = Tmp3dFace->x1; P1[1] = Tmp3dFace->y1; P1[2] = Tmp3dFace->z1;
      P2[0] = Tmp3dFace->x2; P2[1] = Tmp3dFace->y2; P2[2] = Tmp3dFace->z2;
      P3[0] = Tmp3dFace->x3; P3[1] = Tmp3dFace->y3; P3[2] = Tmp3dFace->z3;
      giram_mesh_add_triangle(TmpMesh, P1, P2, P3);

      if ((Tmp3dFace->x3 != Tmp3dFace->x4) ||
          (Tmp3dFace->y3 != Tmp3dFace->y4) ||
          (Tmp3dFace->z3 != Tmp3dFace->z4))
      {

        P1[0] = Tmp3dFace->x1; P1[1] = Tmp3dFace->y1; P1[2] = Tmp3dFace->z1;
        P2[0] = Tmp3dFace->x4; P2[1] = Tmp3dFace->y4; P2[2] = Tmp3dFace->z4;
        P3[0] = Tmp3dFace->x3; P3[1] = Tmp3dFace->y3; P3[2] = Tmp3dFace->z3;
        giram_mesh_add_triangle(TmpMesh, P1, P2, P3);
      }
    }
    if (TmpFrame->FirstObject)
    {
      for (TmpObject = TmpFrame->FirstObject ; 
           TmpObject->Next ; 
           TmpObject = TmpObject->Next)
        ;
      TmpObject->Next = FirstMesh;
    } else
    {
      TmpFrame->FirstObject = FirstMesh;
    }
  }

  /* Lines*/
  for (TmpLine = dat->lines ; TmpLine ; TmpLine=TmpLine->next)
  {
    ObjectStruct *Cylinder;
    Vector Base, Apex;

    Base[0] = TmpLine->x1; Base[1] = TmpLine->y1; Base[2] = TmpLine->z1;
    Apex[0] = TmpLine->x2; Apex[1] = TmpLine->y2; Apex[2] = TmpLine->z2;
    Cylinder = giram_cylinder_new(Base, Apex, 10e-4);
    Cylinder->Next = TmpFrame->FirstObject;
    TmpFrame->FirstObject = Cylinder;
  }

  /* Vertices */
  for (TmpVertex = dat->verts ; TmpVertex ; TmpVertex=TmpVertex->next)
  {
    ObjectStruct *Sphere;
    Vector Center;

    Center[0]=TmpVertex->x; Center[1]=TmpVertex->y; Center[2]=TmpVertex->z;
    Sphere = giram_sphere_new(Center, 10e-4);
    Sphere->Next = TmpFrame->FirstObject;
    TmpFrame->FirstObject = Sphere;
  }
/**************************************************************************/  
  dxf_free_all(dat);
  free(dat);
  for (TmpView=TmpFrame->FirstView ; TmpView ; TmpView=TmpView->Next)
    gtk_widget_queue_draw(TmpView->DrawingArea);
  RebuildCSGTree(NULL, TmpFrame);
}
#endif


typedef struct DXFPrivate
{
  FrameStruct *Frame;
  char *FileName;
  int SelectedOnly;
  GtkWidget *Text;
  int IncludeComment;
} DXFPrivate; 

/*****************************************************************************
*  dxf_save_file
*  Dumps out Giram triangle data in an ASCII (for now) DXF file
******************************************************************************/
int dxf_save_file(FrameStruct *frame, char *fname, int binary, int SelectedOnly,
                  char *Comment)
{
  FILE *fp;
  ObjectStruct *iterObject;
  ObjectSelection *iterSel;

  if (binary)
  {
    fprintf (stderr, "DXF_SAVE: Binary writes not implemented yet.\n");
    return 1;
  }

  if (SelectedOnly)
  {
    if (frame->Selection == NULL)
    {
      fprintf(stderr, "DXF_SAVE: Nothing to write!\n");
      return 1;
    }
  } else
  {
    if (frame->FirstObject == NULL)
    {
      fprintf(stderr, "DXF_SAVE: Nothing to write!\n");
      return 1;
    }
  }
  fp = fopen(fname,"w");

  if (fp == NULL)
  {
    fprintf(stderr, "DXF_SAVE: Couldn't open %s for writing.\n", fname);
    return 1;
  }

  /* Okay, we have an open file, and at least one thing to write to it.
   * Lets start dumping...
   */


  /* First, the header, eschewing all but the ENTITIES section */
  if (!dxf_write_header(fp, binary, Comment))
  {
    fclose(fp);
    return 1;
  }

  if (SelectedOnly)
  {
    for (iterSel = frame->Selection ;
         iterSel->Next != frame->Selection ;
         iterSel = iterSel->Next)
    {
      if (!dxf_write_object(fp, iterSel->Object))
      {
        fclose(fp);
        return 1;
      }
    }
    if (!dxf_write_object(fp, iterSel->Object))
    {
      fclose(fp);
      return 1;
    }
  } else
  {
    for (iterObject=frame->FirstObject ; 
         iterObject ; 
         iterObject=iterObject->Next)
    {
      if (!dxf_write_object(fp, iterObject))
      {
        fclose(fp);
        return 1;
      }
    }
  }

  if (!dxf_write_end(fp, binary))
  {
    fclose(fp);
    return 1;
  }

  fclose (fp);
  return 0;
}

/*****************************************************************************
*  dxf_write_header
******************************************************************************/
int dxf_write_header(FILE *fp, int bin, char *Comment)
{
  int strpos = 0;
  char CommentLine[256];
#if 0
  if (bin) ; /* Nothing for now. */
#endif
  /* The blank HEADER section */
  fprintf(fp, "0\nSECTION\n2\nHEADER\n0\nENDSEC\n");

  /* Comments out, 256 Bytes per line */
  if (Comment)
  {
    while (Comment[strpos])
    {
      int i;
      for (i=0; 
           (Comment[strpos]) && (Comment[strpos]!='\n') && (i < 255) ;
           i++, strpos++)
      {
        CommentLine[i] = Comment[strpos];
      }
      if (i<255) strpos++;
      CommentLine[i] = 0;
      fprintf(fp, "999\n%s\n", CommentLine);
    }
  }

  /* The blank HEADER section */
  fprintf(fp, "0\nSECTION\n2\nHEADER\n0\nENDSEC\n");

  /* Start the ENTITIES section */
  fprintf(fp, "0\nSECTION\n2\nENTITIES\n");

  return TRUE;
}

/*****************************************************************************
*  dxf_write_object
******************************************************************************/
int dxf_write_object(FILE *fp, ObjectStruct *obj)
{
  TriangleListStruct *tri;

  if ((tri=obj->FirstTriangle) == NULL)
    return TRUE;   /* Nothing to write */

  for ( ; tri ; tri=tri->Next)
  {
    /* The entity.. always a 3dface for now */
    fprintf(fp, "0\n3DFACE\n");

    /* The layer, blank until Giram supports it */
    fprintf(fp, "8\nGiramLayer\n");

    fprintf(fp, "10\n%f\n", tri->P1[0]);
    fprintf(fp, "20\n%f\n", tri->P1[1]);
    fprintf(fp, "30\n%f\n", tri->P1[2]);

    fprintf(fp, "11\n%f\n", tri->P2[0]);
    fprintf(fp, "21\n%f\n", tri->P2[1]);
    fprintf(fp, "31\n%f\n", tri->P2[2]);

    fprintf(fp, "12\n%f\n", tri->P3[0]);
    fprintf(fp, "22\n%f\n", tri->P3[1]);
    fprintf(fp, "32\n%f\n", tri->P3[2]);

    fprintf(fp, "13\n%f\n", tri->P3[0]);
    fprintf(fp, "23\n%f\n", tri->P3[1]);
    fprintf(fp, "33\n%f\n", tri->P3[2]);
  }
  return TRUE;
}

/*****************************************************************************
*  dxf_write_end
******************************************************************************/
int dxf_write_end(FILE *fp, int bin)
{
#if 0 
  if (bin) ; /* GNDN */
#endif
  fprintf(fp, "0\nENDSEC\n0\nEOF\n");

  return TRUE;
}

/*****************************************************************************
*  SaveDXFOK
******************************************************************************/
void SaveDXFOK(GtkObject *Dialog)
{
  DXFPrivate *DXFData;

  DXFData = gtk_object_get_data(Dialog, "DXFData");

  if (DXFData->IncludeComment)
    dxf_save_file(DXFData->Frame, DXFData->FileName, 0, DXFData->SelectedOnly,
                  gtk_editable_get_chars(GTK_EDITABLE(DXFData->Text),0,-1) );
  else
    dxf_save_file(DXFData->Frame, DXFData->FileName, 0, DXFData->SelectedOnly,
                  NULL);
  g_free(DXFData);
}

/*****************************************************************************
*  ToggleDXFIncludeComment
******************************************************************************/
void ToggleDXFIncludeComment(GtkObject *Dialog)
{
  DXFPrivate *DXFData;

  DXFData = gtk_object_get_data(Dialog, "DXFData");
  DXFData->IncludeComment = !DXFData->IncludeComment;
}

/*****************************************************************************
*  SaveDXFDialog
******************************************************************************/
void SaveDXFDialog(FrameStruct *Frame, char *FileName, int SelectedOnly)
{
  GtkWidget *Dialog;
  GtkWidget *Button;
  GtkWidget *CheckButton;
  GtkWidget *frame;
  GtkWidget *Table;
  GtkWidget *Text;
  GtkWidget *Vscrollbar;
  char *Comment = "This file was Generated by Giram-" VERSION "\n"
                  "If you experience any error, you can reach the\n"
                  "author of the Giram's DXF exporter at <adoyle@rimmer.viewsnet.com>\n";
  DXFPrivate *DXFData;

  DXFData = g_malloc(sizeof(DXFPrivate));
  DXFData->Frame = Frame;
  DXFData->FileName = g_strdup(FileName);
  
  DXFData->SelectedOnly = SelectedOnly;
  /* The Dialog Box for the Optionnal Comment */
  Dialog = gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(Dialog), "Save as Jpeg");
  gtk_window_position(GTK_WINDOW(Dialog), GTK_WIN_POS_MOUSE);
  gtk_object_set_data(GTK_OBJECT(Dialog), "DXFData", DXFData);
  
  /* The "OK" Button */
  Button = gtk_button_new_with_label(_("OK"));
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->action_area), Button, TRUE, TRUE, 0);
  gtk_signal_connect_object(GTK_OBJECT(Button), "clicked",
                            GTK_SIGNAL_FUNC(SaveDXFOK),
                            GTK_OBJECT(Dialog));
  gtk_signal_connect_object(GTK_OBJECT(Button), "clicked",
                            GTK_SIGNAL_FUNC(gtk_widget_destroy),
                            GTK_OBJECT(Dialog));
  GTK_WIDGET_SET_FLAGS(Button, GTK_CAN_DEFAULT);
  gtk_widget_grab_default(Button);
  gtk_widget_show(Button);

  /* The Cancel Button */
  Button = gtk_button_new_with_label(_("Cancel"));
  GTK_WIDGET_SET_FLAGS(Button, GTK_CAN_DEFAULT);
  gtk_signal_connect_object(GTK_OBJECT(Button), "clicked",
                            (GtkSignalFunc)gtk_widget_destroy,
                            GTK_OBJECT(Dialog));
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->action_area), Button, TRUE, TRUE, 0);
  gtk_widget_show(Button);
  
  /* The check button for enabling/disabling comment */
  CheckButton = gtk_check_button_new_with_label(_("Save with comment"));
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->vbox), CheckButton, FALSE, FALSE, 2);
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(CheckButton), TRUE);
  DXFData->IncludeComment = TRUE;
  gtk_signal_connect_object(GTK_OBJECT(CheckButton), "toggled",
                            GTK_SIGNAL_FUNC(ToggleDXFIncludeComment),
                            GTK_OBJECT(Dialog));
  gtk_widget_show(CheckButton);


  frame = gtk_frame_new("Comments:");
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);
  gtk_container_border_width(GTK_CONTAINER(frame), 10);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(Dialog)->vbox), frame, TRUE, TRUE, 0);
  gtk_widget_show(frame);

  Table = gtk_table_new(1, 1, FALSE);
  gtk_container_border_width(GTK_CONTAINER(Table), 10);
  gtk_container_add(GTK_CONTAINER(frame), Table);
  gtk_widget_show(Table);

  Text = gtk_text_new(NULL, NULL);
  gtk_text_set_editable(GTK_TEXT(Text), TRUE);
  gtk_text_insert(GTK_TEXT(Text), NULL, NULL, NULL, Comment, -1);
  gtk_table_attach(GTK_TABLE(Table), Text, 0, 1, 0, 1,
                   GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                   GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
  DXFData->Text = Text;
  gtk_widget_show(Text);

  /* Add a vertical scrollbar to the GtkText widget */
  Vscrollbar = gtk_vscrollbar_new(GTK_TEXT(Text)->vadj);
  gtk_table_attach(GTK_TABLE(Table), Vscrollbar, 1, 2, 0, 1,
                   GTK_FILL, GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);
  gtk_widget_show(Vscrollbar);
 
  gtk_widget_show(Dialog);
  
}

/*****************************************************************************
*  SaveDXF
******************************************************************************/
void SaveDXF(GtkWidget *w, GtkFileSelection *FS)
{
  FrameStruct *LocalFrame;
  char *FileName;

  FileName = gtk_file_selection_get_filename(GTK_FILE_SELECTION(FS));
  LocalFrame = (FrameStruct *)gtk_object_get_data(GTK_OBJECT(FS), "frame");
  SaveDXFDialog(LocalFrame, FileName, FALSE);
}

/*****************************************************************************
*  SaveSelectedDXF
******************************************************************************/
void SaveSelectedDXF(GtkWidget *w, GtkFileSelection *FS)
{
  FrameStruct *LocalFrame;
  char *FileName;

  FileName = gtk_file_selection_get_filename(GTK_FILE_SELECTION(FS));
  LocalFrame = (FrameStruct *)gtk_object_get_data(GTK_OBJECT(FS), "frame");
  SaveDXFDialog(LocalFrame, FileName, TRUE);
}


#ifdef TEST_DXF
int main(int argc, char *argv[])
{ 
	//return (dxf_parse(argv[1])); 

	return 0;
}
#endif
