/* -*- Mode: C++; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*- */
/*===========================================================================
 * 
 * Project : Freyja
 * Author  : Terry 'Mongoose' Hendrix II
 * Website : http://icculus.org/freyja/
 * Email   : mongooseichiban@gmail.com
 * Object  : Nif5
 * License : No use w/o permission (C) 2007 Mongoose
 * Comments: Gamebryo Model (Oblivion) plugin
 *
 *-- History ------------------------------------------------ 
 *
 * 2007.07.04:
 * Mongoose - Created, a new stripped down Nif5 plugin KISS
 ==========================================================================*/

#include "Nif5.h"

// Toggle debug printing for reverse engineering
#define PRINT printf
// #define PRINT(...)



////////////////////////////////////////////////////////////
// Constructors
////////////////////////////////////////////////////////////

Nif5::Nif5() 
	: mHeaderSz(98) /* actually variable lenght, but 98 in test model */
{
	mHeader[0] = '\0';
}


Nif5::~Nif5()
{
}


////////////////////////////////////////////////////////////
// Public Accessors
////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////
// Public Mutators
////////////////////////////////////////////////////////////

bool Nif5::loadModel(const char *filename)
{
	SystemIO::FileReader r;

	if (r.Open(filename) == false)
		return false;

	/* Get header and check to see if this is a version match. */
	r.ReadString(38, mHeader);
	mHeader[38] = 0;

	if ( strncmp(NIF5_HEADER_START, mHeader, 38) )
		return false;

	PRINT("Header magic '%s'\n", mHeader);

	// 'header strings' end with 0x0A

	// Likely part of header string padding: 0A05 0000 1401 0B00 0000
	{
		for (unsigned int i = 0, u; i < 5; ++i)
		{
			int u = r.ReadInt16();
			PRINT("unknown_static%i %i\n", i, u);
		}
	}

	{
		for (unsigned int i = 0, u; i < 2; ++i)
		{
			int u = r.ReadInt32();
			PRINT("unknown%i %i\n", i, u);
		}

		
		int u = r.ReadInt16();
		PRINT("unknown_pad? %i\n", u);
	}
	
	{
		int len = r.ReadByte();
		char s[len+1];
		r.ReadString(len, s);
		s[len] = 0;
		PRINT("comment %i '%s'\n", len, s);
	}
	
	Vector<NiNode *> ninodes;
	Chunk chunk;

	{
		int i = r.ReadInt16();;
		PRINT("?: %i\n", i);
	}

	int count = r.ReadInt16();;
	PRINT("chunk count: %i\n", count);
	chunk.count = 0;



	while (chunk.ReadHeader(r) && !r.IsEndOfFile() && chunk.count < count)
	{
		PRINT("Ni Chunk: #%i. '%s'\n", chunk.count, chunk.name);
		++chunk.count;

		if (!strncmp(chunk.name, "NiNode", 6))
		{
			NiNode *ninode = new NiNode();			
			ninodes.push_back(ninode);
			//ninode->readChunk(r);
		}
		else if (!strncmp(chunk.name, "NiTextKeyExtraData", 18))
		{
			NiTextKeyExtraData *chunk = new NiTextKeyExtraData();			
			//mNiTextKeyExtraData.push_back(chunk);
			//chunk->readChunk(r);
		}
		else if (!strncmp(chunk.name, "NiTriShapeData", 14))
		{
			NiTriShapeData *chunk = new NiTriShapeData();			
			mNiTriShapeData.push_back(chunk);
			//chunk->readChunk(r);
		}
		else if (!strncmp(chunk.name, "NiTriShape", 10))
		{
			NiTriShape *nitrishape = new NiTriShape();			
			//nitrishapes.push_back(nitrishape);
			//nitrishape->readChunk(r);
		}
		else if (!strncmp(chunk.name, "NiGeomMorpherController", 23))
		{
			 NiGeomMorpherController *chunk = new NiGeomMorpherController();
			//mNiTextKeyExtraData.push_back(chunk);
			//chunk->readChunk(r);
		}
		else if (!strncmp(chunk.name, "NiMorphData", 11))
		{
			 NiMorphData *chunk = new NiMorphData();			
			//mNiTextKeyExtraData.push_back(chunk);
			//chunk->readChunk(r);
		}
		else if (!strncmp(chunk.name, "NiTexturingProperty", 19))
		{
			 NiTexturingProperty *chunk = new NiTexturingProperty();			
			//mNiTextKeyExtraData.push_back(chunk);
			//chunk->readChunk(r);
		}
		else if (!strncmp(chunk.name, "NiSourceTexture", 15))
		{
			 NiSourceTexture *chunk = new NiSourceTexture();			
			//mNiTextKeyExtraData.push_back(chunk);
			//chunk->readChunk(r);
		}
		else if (!strncmp(chunk.name, "NiMaterialProperty", 18))
		{
			 NiMaterialProperty *chunk = new NiMaterialProperty();			
			//mNiTextKeyExtraData.push_back(chunk);
			//chunk->readChunk(r);
		}
		else
		{
			PRINT("Ni Chunk: '%s', %i @ %li not implemented\n",
				   chunk.name, chunk.len, r.GetOffset());

			//r.Close();
			//break;
		}
	}

	// Ghost.nif

	// @ 749

	PRINT("\n");

	{
		r.SetOffset(749);
		unsigned int len = r.ReadInt32();
		char s[len+1];
		r.ReadString(len, s);
		s[len] = 0;
		PRINT("node: '%s'\n", s);

		PRINT("id? %i\n", r.ReadInt32() );
		PRINT("parent? %i\n", r.ReadInt32() );
		
	}


	// 14232		@ 308142
	// float[]		@ 308146
	if (0)
	{
		r.SetOffset(308142);
		unsigned int n = r.ReadInt32() / 3;

		PRINT("%i\n", n);

		for (unsigned int i = 0; i < n; ++i)
		{
			PRINT("%i. %f %f %f\n", i, r.ReadFloat32(), r.ReadFloat32(), r.ReadFloat32());
		}
	}
	

	// vertices[]?	@ 

	// uvs[]?		@ 

	// faces[]?		@ 341389 { short[3] : a, b, c }

	// 593 short	@ 343999
	// weights[]?	@ 344001  { short id, float weight }
	// weights[]?	@ 347629  { short id, float weight }  

	r.Close();

	return true;
}


////////////////////////////////////////////////////////////
// Private Accessors
////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////
// Private Mutators
////////////////////////////////////////////////////////////



////////////////////////////////////////////////////////////
// Plugin hooks
////////////////////////////////////////////////////////////

#ifdef FREYJA_PLUGINS
#include <freyja/ModelABI.h>
#include <freyja/PluginABI.h>
#include <freyja/BoneABI.h>
#include <freyja/SkeletonABI.h>
#include <freyja/TextureABI.h>
#include <freyja/LegacyABI.h>
#include <freyja/MeshABI.h>
#include <freyja/freyja.h>
#include <mstl/Vector.h>

using namespace mstl;

extern "C" {

	int freyja_model__nif5_check(char *filename);
	int freyja_model__nif5_import(char *filename);
	int freyja_model__nif5_export(char *filename);
	int import_model(char *filename);
	void freyja_init();
}


void freyja_init()
{
	freyjaPluginName1s("nif5");
	freyjaPluginDescription1s("NetImmerse 5 Model (*.nif)");
	freyjaPluginAddExtention1s("*.nif");
	freyjaPluginImport1i(FREYJA_PLUGIN_MESH);
	freyjaPluginExport1i(FREYJA_PLUGIN_NONE);
}


int import_model(char *filename)
{
	if (!freyja_model__nif5_check(filename))
		return freyja_model__nif5_import(filename);

	return -1;
}


int freyja_model__nif5_check(char *filename)
{	
	SystemIO::FileReader r;
	char header[48];
	unsigned int headerSz = 48;

	if (r.Open(filename) == false)
		return -1;

	/* Get header and check to see if this is a version match. */
	r.ReadString(headerSz, header);
	r.Close();

	if (strncmp(NIF5_HEADER_START, header, 38))
		return -1;

	return 0;
}


int freyja_model__nif5_import(char *filename)
{
	Vector<unsigned int> vertices;
	Vector<unsigned int> texcoords;
	Nif5 Nif5;
	long i, j, idx;


	if (Nif5.loadModel(filename) == false)
		return -1;

	freyjaBegin(FREYJA_MODEL);


	for (i = Nif5.mNiTriShapeData.begin(); i < (int)Nif5.mNiTriShapeData.end(); ++i)
	{
		Nif5::NiTriShapeData *data = Nif5.mNiTriShapeData[i];

		if (!data)
			continue;

		vertices.clear();
		texcoords.clear();

		freyjaBegin(FREYJA_MESH);
		freyjaBegin(FREYJA_VERTEX_GROUP);

		for (j = 0; j < data->num_vertices; ++j)
		{
			idx = freyjaVertexCreate3f(data->coordinates[j].x,
								 data->coordinates[j].z,
								 data->coordinates[j].y);

			freyjaVertexNormal3f(idx,
								 data->normals[j].x,
								 data->normals[j].z,
								 data->normals[j].y);

			if (data->num_uv_sets > 0)
			{
				freyjaTexCoord2f(idx,
								 data->uv_set[0].textureinfo[j].u,
								 data->uv_set[0].textureinfo[j].v);
			}

			vertices.push_back(idx);
		}

		freyjaEnd(); // FREYJA_VERTEX_GROUP

		for (j = 0; j < data->num_triangles; ++j)
		{
			freyjaBegin(FREYJA_POLYGON);
			freyjaPolygonMaterial1i(0);
			freyjaPolygonVertex1i(vertices[data->triangles[j].v[0]]);
			freyjaPolygonVertex1i(vertices[data->triangles[j].v[1]]);
			freyjaPolygonVertex1i(vertices[data->triangles[j].v[2]]);
			freyjaEnd(); // FREYJA_POLYGON
		}

		freyjaEnd(); // FREYJA_MESH		
	}

	freyjaEnd(); // FREYJA_MODEL

	return 0;
}


int freyja_model__nif5_export(char *filename)
{
	freyjaPrintError("NIF5_export> ERROR: Not implemented.\n");
	return -1;
}
#endif


////////////////////////////////////////////////////////////
// Unit Test code
////////////////////////////////////////////////////////////

#ifdef UNIT_TEST_Nif5
int runNif5UnitTest(int argc, char *argv[])
{
	Nif5 Nif5;

	if (Nif5.loadModel(argv[1]) == false)
		return -1;

	return 0;
}


int main(int argc, char *argv[])
{
	printf("[Nif5 class test]\n");

	return runNif5UnitTest(argc, argv);
}
#endif
